package madynes.flowoid;

import madynes.flowoid.exporter.R;
import android.content.Context;
import android.location.Location;
import android.location.LocationListener;
import android.location.LocationManager;
import android.os.Bundle;
import android.os.Handler;
import android.os.HandlerThread;
import android.util.Log;

public class LocationFinder implements LocationListener {
	private Location location;
	private LocationManager locationManager;
	private static String providerName;
	private static LocationFinder instance;
	private static int minLocationInterval = 0;
	private static int minLocationDistance = 0;
	private HandlerThread passiveThread;
	private Handler passiveHandler;
	private Runnable passiveRun;
	private Context context;

	/**
	 * Returns the LocationFinder singleton. 
	 * 
	 * @param context an Android context
	 * @return singleton
	 */
	public static LocationFinder getInstance(Context context) {
		synchronized (LocationFinder.class) {
			if(instance == null) {
				instance = new LocationFinder(context);
			}			
		}

		return instance;
	}
	
	/**
	 * Sets the minimum time interval between location updates, in seconds.
	 * 
	 * @param interval interval in seconds
	 */
	public static void setMinLocationInterval(int interval) {
		minLocationInterval = interval;
	}
	
	/**
	 * Returns the minimum time interval between location updates, in seconds.
	 * 
	 * @return interval in seconds
	 */
	public static int getMinLocationInterval() {
		return minLocationInterval;
	}
	
	/**
	 * Sets minimum distance between location updates, in meters.
	 * 
	 * @param interval distance in meters
	 */
	public static void setMinLocationDistance(int distance) {
		minLocationDistance = distance;
	}
	
	/**
	 * Returns the minimum distance between location updates, in meters.
	 * 
	 * @return distance in meters
	 */
	public static int getMinLocationDistance() {
		return minLocationDistance;
	}
	
	/**
	 * Defines a new LocationFinder updated each time the device is moved.
	 * 
	 * @param context an Android context
	 */
	private LocationFinder(Context context) {
	    locationManager = (LocationManager) context.getSystemService(Context.LOCATION_SERVICE);
	    this.context = context;
	}
	
	/**
	 * Requests updates.
	 */
	public void start() {
	    String provider = PreferencesActivity.getLocationProvider(context);

    	location = locationManager.getLastKnownLocation(LocationManager.PASSIVE_PROVIDER);
	    
	    if(provider == LocationManager.GPS_PROVIDER) {
	    	providerName = context.getString(R.string.GPS);
	    } else if(provider == LocationManager.NETWORK_PROVIDER) {
	    	providerName = context.getString(R.string.Network);
	    } else if(provider == LocationManager.PASSIVE_PROVIDER) {
	    	providerName = context.getString(R.string.Passive);
	    } else
	    	providerName = context.getString(R.string.Unknown);
	    
	    if(provider == LocationManager.PASSIVE_PROVIDER) {
			if(passiveHandler != null) {
				passiveHandler.removeCallbacks(passiveRun);
				passiveThread.quit();
				passiveHandler = null;
			}
			
	    	passiveThread = new HandlerThread("LocationThread");
	    	passiveThread.start();
	    	passiveHandler = new Handler(passiveThread.getLooper());
	    	
	    	passiveRun = new Runnable() {
				@Override
				public void run() {
					location = locationManager.getLastKnownLocation(LocationManager.PASSIVE_PROVIDER);

					Log.d("THREAD", "Checking passive location");
					
					if(passiveHandler != null) {
						passiveHandler.postDelayed(this, minLocationInterval * 1000);
					}
				}
	    	};

	    	passiveHandler.post(passiveRun);
	    	
	    } else {
		    locationManager.requestLocationUpdates(provider, minLocationInterval * 1000, minLocationDistance, this);
	    }
	}
	
	/**
	 * Declines updates.
	 */
	public void stop() {
		locationManager.removeUpdates(this);
		
		if(passiveHandler != null) {
			passiveHandler.removeCallbacks(passiveRun);
			passiveThread.quit();
			passiveHandler = null;
		}
	}
	
	/**
	 * Returns directly the Location object.
	 * 
	 * @return Current Location
	 */
	public Location getLocation() {
		return location;
	}
	
	/**
	 * Returns the last known latitude.
	 * 
	 * @return Last known latitude
	 */
	public double getLatitude() {
		return (location != null) ? location.getLatitude() : -1;
	}
	
	/**
	 * Returns the last known longitude.
	 * 
	 * @return Last known longitude
	 */
	public double getLongitude() {
		return (location != null) ? location.getLongitude() : -1;
	}
	
	/**
	 * Returns the last known altitude.
	 * 
	 * @return Last known altitude
	 */
	public double getAltitude() {
		return (location != null) ? location.getAltitude() : -1;
	}
	
	/**
	 * Returns the date of the last known location.
	 * 
	 * @return date of the last known location (seconds from January, 1st 1970)
	 */
	public int getTimestamp() {
		return (int) ((location != null) ? location.getTime() / 1000 : -1);
	}
	
	/**
	 * Returns the provider used in human format.
	 * 
	 * @return location provider name
	 */
	public static String getProviderName() {
		return providerName;
	}
	
	/**
	 * Updates location each time the device is moved.
	 */
	@Override
	public void onLocationChanged(Location location) {
		this.location = location;
	}

	@Override
	public void onProviderDisabled(String provider) {}

	@Override
	public void onProviderEnabled(String provider) {}

	@Override
	public void onStatusChanged(String provider, int status, Bundle extras) {}
}
