package madynes.flowoid;

import madynes.flowoid.exporter.ExporterService;
import madynes.flowoid.exporter.R;
import android.content.Context;
import android.content.SharedPreferences;
import android.content.SharedPreferences.OnSharedPreferenceChangeListener;
import android.location.LocationManager;
import android.os.Bundle;
import android.preference.PreferenceActivity;
import android.preference.PreferenceManager;
import android.widget.Toast;

public class PreferencesActivity extends PreferenceActivity implements OnSharedPreferenceChangeListener {
	final static private int LOC_PASSIVE = 0;
	final static private int LOC_NETWORK = 1;
	final static private int LOC_GPS = 2;
	
    @SuppressWarnings("deprecation")
	@Override
    public void onCreate(Bundle savedInstanceState) {
        super.onCreate(savedInstanceState);
        
        // TODO: Find no deprecated methods
        addPreferencesFromResource(R.layout.preferences);
        PreferenceManager.setDefaultValues(PreferencesActivity.this, R.layout.preferences, false);
        getPreferenceScreen().getSharedPreferences().registerOnSharedPreferenceChangeListener(this);
    }
    
    /**
     * Returns preference value (type boolean).
     * 
     * @param context an Android context
     * @param pref preference name
     * @param def default value
     * @return preference value
     */
    public static boolean getBooleanPreference(Context context, String pref, boolean def) {
    	SharedPreferences prefs = PreferenceManager.getDefaultSharedPreferences(context);

    	return prefs.getBoolean(pref, def);
    }
    
    /**
     * Returns preference value (type integer).
     * 
     * @param context an Android context
     * @param pref preference name
     * @param def default value
     * @return preference value
     */
    public static int getIntPreference(Context context, String pref, int def) {
    	SharedPreferences prefs = PreferenceManager.getDefaultSharedPreferences(context);

    	return Integer.valueOf(prefs.getString(pref, String.valueOf(def)));
    }
    
    /**
     * Returns as pcap format the filters defined by the user for the IP Driver. 
     * 
     * @param context an Android context
     * @return filters as pcap format
     */
    public static String getFiltersIP(Context context) {
    	SharedPreferences prefs = PreferenceManager.getDefaultSharedPreferences(context);
    	final String[] protos = new String[] { "protoIpIcmp", "protoIpIcmp6", "protoIpUdp", "protoIpTcp" };
    	StringBuffer filters = new StringBuffer();

    	if(!prefs.getBoolean("enableDriverIP", true)) {
    		return null;
    	}
    	
    	for(String proto : protos) {
    		if(prefs.getBoolean("enableDriverIP", true)) {
    			filters.append(" or ");
    			filters.append(proto.replace("protoIp", "").toLowerCase());
    		}
    	}
    	
    	boolean enableIPv4 = prefs.getBoolean("enableIPv4", true);
    	boolean enableIPv6 = prefs.getBoolean("enableIPv6", true);
    	
    	if(filters.length() != 0 && (enableIPv4 || enableIPv6)) {
    		filters.append(" and (");
    		
    		if(enableIPv4) {
    			filters.append("ip");
    		}

    		if(enableIPv4 && enableIPv6) {
    			filters.append(" or ");
    		}

    		if(enableIPv6) {
        		filters.append("ip6");    		
    		}
    		
    		filters.append(")");
    		
    	} else {
    		return null;
    	}
    	
    	return filters.substring(4);
    }
    
    /**
     * Returns the Location Provider chosen by the user.
     * 
     * @param context an Android context
     * @return Location Provider
     */
    public static String getLocationProvider(Context context) {
    	SharedPreferences prefs = PreferenceManager.getDefaultSharedPreferences(context);
    	
    	switch(Integer.parseInt(prefs.getString("locationProvider", "" + LOC_PASSIVE))) {
    	case LOC_NETWORK:
    		return LocationManager.NETWORK_PROVIDER;
    		
    	case LOC_GPS:
    		return LocationManager.GPS_PROVIDER;
    		
    	case LOC_PASSIVE:
    	default:
    		return LocationManager.PASSIVE_PROVIDER;
    	}
    }
  
    /**
     * Configures the Exporter Service preferences with the user choices.
     * Default values are here and in layout/preferences.xml.
     * 
     * @param context an Android context
     */
    public static void configureExporter(Context context) {
    	SharedPreferences prefs = PreferenceManager.getDefaultSharedPreferences(context);
    	
    	ExporterService.setPrefCheckDriversInterval(Integer.parseInt(prefs.getString("checkDriversInterval", "60")));
    	ExporterService.setPrefCheckToExportInterval(Integer.parseInt(prefs.getString("checkToExportInterval", "120")));
    	ExporterService.setPrefCollectorAddr(prefs.getString("collectorAddr", "152.81.67.203"));
    	ExporterService.setPrefCollectorPort(Integer.parseInt(prefs.getString("collectorPort", "9995")));
    	ExporterService.setPrefDriversPort(Integer.parseInt(prefs.getString("driversPort", "12349")));
    	ExporterService.setPrefMaxHeadersSize(Integer.parseInt(prefs.getString("maxHeadersSize", "1024")));
    	ExporterService.setPrefMaxNetflowsToTrack(Integer.parseInt(prefs.getString("maxNetflowsToTrack", "2000")));
    	ExporterService.setPrefMinRecordsByFlowset(Integer.parseInt(prefs.getString("minRecordsByFlowset", "5")));
    	ExporterService.setPrefNetflowTimeout(Integer.parseInt(prefs.getString("netflowTimeout", "300")));
    	ExporterService.setPrefPacketTimeout(Integer.parseInt(prefs.getString("packetTimeout", "600")));
    	ExporterService.setPrefSendTemplatesEveryNPackets(Integer.parseInt(prefs.getString("sendTemplatesEveryNPackets", "10")));
    	ExporterService.setPrefTimeBeforeNetFlowEnd(Integer.parseInt(prefs.getString("timeBeforeNetFlowEnd", "60")));
    	ExporterService.setPrefMaxExportPacketSize(Integer.parseInt(prefs.getString("maxExportPacketSize", "1024")));
    }

	@Override
	public void onSharedPreferenceChanged(SharedPreferences sharedPreferences, String key) {
		if(ExporterService.isRunning()) {
			Toast.makeText(this, "Restart the probe for the changes take effect", Toast.LENGTH_SHORT).show();
		}
	}
}