package madynes.flowoid.exporterImpl;

import java.util.ArrayList;

import android.content.IntentFilter;
import android.net.ConnectivityManager;
import android.util.Log;

import madynes.flowoid.LocationFinder;
import madynes.flowoid.PreferencesActivity;
import madynes.flowoid.Connectivity;
import madynes.flowoid.exporter.DataRecord;
import madynes.flowoid.exporter.ExporterService;
import madynes.flowoid.exporter.TemplateField;
import madynes.flowoid.exporter.TemplateField.InadequateLengthException;
import madynes.flowoid.exporter.TemplateFlowSet.TemplateNotFoundException;
import madynes.flowoid.exporter.TemplateRecord;
import madynes.flowoid.exporter.TemplateRecord.AlreadyUsedIDException;
import madynes.flowoid.exporter.TemplateRecord.InadequateIDException;
import madynes.flowoid.exporterImpl.FieldIPProtocol.IPProtocol;
import madynes.flowoid.exporterImpl.FieldIPVersion.IPVersion;
import edu.huji.cs.netutils.parse.IPFactory;
import edu.huji.cs.netutils.parse.IPv4Packet;
import edu.huji.cs.netutils.parse.IPv6Packet;

public class Exporter extends ExporterService {
	private Connectivity connectivity;
	private LocationFinder locationFinder;
	private boolean enableIPv4, enableIPv6, protoIpIcmp, protoIpIcmp6, protoIpUdp, protoIpTcp;

	private final int TPL_IPV4_ICMP_ID = 4094;
	private final int TPL_IPV4_UDP_ID = 4095;
	private final int TPL_IPV4_TCP_ID = 4096;
	private final int TPL_IPV6_ICMPV6_ID = 8190;
	private final int TPL_IPV6_UDP_ID = 8191;
	private final int TPL_IPV6_TCP_ID = 8192;
	
	/**
	 * Defines NetFlow v9 templates.
	 */
	public void registerTemplates() throws AlreadyUsedIDException, InadequateIDException, InadequateLengthException {
		ArrayList<TemplateRecord> templateIPRecords = new ArrayList<TemplateRecord>();
		
		enableIPv4 = PreferencesActivity.getBooleanPreference(this, "enableIPv4", true);
		enableIPv6 = PreferencesActivity.getBooleanPreference(this, "enableIPv6", true);
		protoIpIcmp = PreferencesActivity.getBooleanPreference(this, "protoIpIcmp", true);
		protoIpIcmp6 = PreferencesActivity.getBooleanPreference(this, "protoIpIcmp6", true);
		protoIpUdp = PreferencesActivity.getBooleanPreference(this, "protoIpUdp", true);
		protoIpTcp = PreferencesActivity.getBooleanPreference(this, "protoIpTcp", true);

		// Common IP templates
		TemplateField ipVersion = new TemplateField(TemplateField.TypeDefinition.IP_PROTOCOL_VERSION);
		TemplateField ipProtocol = new TemplateField(TemplateField.TypeDefinition.PROTOCOL);
        TemplateField deviceID = new TemplateField(TemplateField.TypeDefinition.MADYNES_DEVICE_ID);
        TemplateField length = new TemplateField(TemplateField.TypeDefinition.IN_BYTES, 4);
        TemplateField tcpOrUdpPortSrc = new TemplateField(TemplateField.TypeDefinition.L4_SRC_PORT);
        TemplateField tcpOrUdpPortDst = new TemplateField(TemplateField.TypeDefinition.L4_DST_PORT);
        TemplateField tcpFlags = new TemplateField(TemplateField.TypeDefinition.TCP_FLAGS);
        TemplateField firstSwitched = new TemplateField(TemplateField.TypeDefinition.FIRST_SWITCHED);
        TemplateField lastSwitched = new TemplateField(TemplateField.TypeDefinition.LAST_SWITCHED);
        TemplateField packets = new TemplateField(TemplateField.TypeDefinition.IN_PKTS, 4);
        TemplateField locAlgo = new TemplateField(TemplateField.TypeDefinition.MADYNES_LOC_ALGO);
        TemplateField locTimestamp = new TemplateField(TemplateField.TypeDefinition.MADYNES_LOC_TIMESTAMP);
        TemplateField locLatInt = new TemplateField(TemplateField.TypeDefinition.MADYNES_LOC_LAT_INT);
        TemplateField locLatDec = new TemplateField(TemplateField.TypeDefinition.MADYNES_LOC_LAT_DEC);
        TemplateField locLongInt = new TemplateField(TemplateField.TypeDefinition.MADYNES_LOC_LONG_INT);
        TemplateField locLongDec = new TemplateField(TemplateField.TypeDefinition.MADYNES_LOC_LONG_DEC);
        TemplateField appName = new TemplateField(TemplateField.TypeDefinition.MADYNES_APP_NAME, 50);

        /* IPv4 */
        
        if(enableIPv4) {
	        TemplateField ipv4AddrSrc = new TemplateField(TemplateField.TypeDefinition.IPV4_SRC_ADDR);
	        TemplateField ipv4AddrDst = new TemplateField(TemplateField.TypeDefinition.IPV4_DST_ADDR);
        	
        	if(protoIpTcp) {
		        // IPv4 TCP templates
		        TemplateRecord ipv4TcpTemplate = new TemplateRecord(TPL_IPV4_TCP_ID);
		        
		        ipv4TcpTemplate.addTemplateField(ipVersion);
		        ipv4TcpTemplate.addTemplateField(ipv4AddrSrc);
		        ipv4TcpTemplate.addTemplateField(ipv4AddrDst);
		        ipv4TcpTemplate.addTemplateField(length);
		    	ipv4TcpTemplate.addTemplateField(ipProtocol);
		        ipv4TcpTemplate.addTemplateField(tcpOrUdpPortSrc);
		        ipv4TcpTemplate.addTemplateField(tcpOrUdpPortDst);
		        ipv4TcpTemplate.addTemplateField(tcpFlags);
		        ipv4TcpTemplate.addTemplateField(appName);
		        
		        templateIPRecords.add(ipv4TcpTemplate);
        	}
        	
        	if(protoIpUdp) {
		        // IPv4 UDP templates
		        TemplateRecord ipv4UdpTemplate = new TemplateRecord(TPL_IPV4_UDP_ID);
		        
		        ipv4UdpTemplate.addTemplateField(ipVersion);
		        ipv4UdpTemplate.addTemplateField(ipv4AddrSrc);
		        ipv4UdpTemplate.addTemplateField(ipv4AddrDst);
		        ipv4UdpTemplate.addTemplateField(length);
		    	ipv4UdpTemplate.addTemplateField(ipProtocol);
		        ipv4UdpTemplate.addTemplateField(tcpOrUdpPortSrc);
		        ipv4UdpTemplate.addTemplateField(tcpOrUdpPortDst);
		        ipv4UdpTemplate.addTemplateField(appName);
		        
		        templateIPRecords.add(ipv4UdpTemplate);
        	}
        	
        	if(protoIpIcmp) {
		        // IPv4 ICMP templates
		        TemplateRecord ipv4IcmpTemplate = new TemplateRecord(TPL_IPV4_ICMP_ID);
		        
		        ipv4IcmpTemplate.addTemplateField(ipVersion);
		        ipv4IcmpTemplate.addTemplateField(ipv4AddrSrc);
		        ipv4IcmpTemplate.addTemplateField(ipv4AddrDst);
		        ipv4IcmpTemplate.addTemplateField(length);
		        ipv4IcmpTemplate.addTemplateField(ipProtocol);
		
		        templateIPRecords.add(ipv4IcmpTemplate);
        	}
        }
        
        /* IPv6 */
        
        if(enableIPv6) {
	        TemplateField ipv6AddrSrc = new TemplateField(TemplateField.TypeDefinition.IPV6_SRC_ADDR);
	        TemplateField ipv6AddrDst = new TemplateField(TemplateField.TypeDefinition.IPV6_DST_ADDR);
        	
        	if(protoIpTcp) {
		        // IPv6 TCP templates
		        TemplateRecord ipv6TcpTemplate = new TemplateRecord(TPL_IPV6_TCP_ID);
		
		        ipv6TcpTemplate.addTemplateField(ipVersion);
		        ipv6TcpTemplate.addTemplateField(ipv6AddrSrc);
		        ipv6TcpTemplate.addTemplateField(ipv6AddrDst);
		        ipv6TcpTemplate.addTemplateField(length);
		        ipv6TcpTemplate.addTemplateField(ipProtocol);
		        ipv6TcpTemplate.addTemplateField(tcpOrUdpPortSrc);
		        ipv6TcpTemplate.addTemplateField(tcpOrUdpPortDst);
		        ipv6TcpTemplate.addTemplateField(tcpFlags);
		        ipv6TcpTemplate.addTemplateField(appName);
		
		        templateIPRecords.add(ipv6TcpTemplate);
        	}
        	
        	if(protoIpUdp) {
		        // IPv6 UDP templates
		        TemplateRecord ipv6UdpTemplate = new TemplateRecord(TPL_IPV6_UDP_ID);
		
		        ipv6UdpTemplate.addTemplateField(ipVersion);
		        ipv6UdpTemplate.addTemplateField(ipv6AddrSrc);
		        ipv6UdpTemplate.addTemplateField(ipv6AddrDst);
		        ipv6UdpTemplate.addTemplateField(length);
		        ipv6UdpTemplate.addTemplateField(ipProtocol);
		        ipv6UdpTemplate.addTemplateField(tcpOrUdpPortSrc);
		        ipv6UdpTemplate.addTemplateField(tcpOrUdpPortDst);
		        ipv6UdpTemplate.addTemplateField(appName);
		
		        templateIPRecords.add(ipv6UdpTemplate);
        	}
        
	       	if(protoIpIcmp6) {
		        // IPv6 ICMPv6 templates
		        TemplateRecord ipv6Icmpv6Template = new TemplateRecord(TPL_IPV6_ICMPV6_ID);
		        
		        ipv6Icmpv6Template.addTemplateField(ipVersion);
		        ipv6Icmpv6Template.addTemplateField(ipv6AddrSrc);
		        ipv6Icmpv6Template.addTemplateField(ipv6AddrDst);
		        ipv6Icmpv6Template.addTemplateField(length);
		        ipv6Icmpv6Template.addTemplateField(ipProtocol);
		
		        templateIPRecords.add(ipv6Icmpv6Template);
	       	}
        }
        
        // Common IP templates
        for(TemplateRecord templateRecord : templateIPRecords) {
        	templateRecord.addTemplateField(firstSwitched);
        	templateRecord.addTemplateField(lastSwitched);
        	templateRecord.addTemplateField(packets);
        	templateRecord.addTemplateField(deviceID);
        	templateRecord.addTemplateField(locAlgo);
        	templateRecord.addTemplateField(locTimestamp);
        	templateRecord.addTemplateField(locLatInt);
        	templateRecord.addTemplateField(locLatDec);
        	templateRecord.addTemplateField(locLongInt);
        	templateRecord.addTemplateField(locLongDec);
        	
        	registerTemplateRecord(templateRecord);
        }
	}
	
	/**
	 * Process every packet sent by any driver.
	 * 
	 * @param headers byte array of the packet headers (layers 2, 3 and 4)
	 * @return false the treatment failed
	 */
	public boolean processPacketFiltering(byte[] headers) throws TemplateNotFoundException {
		DataRecord record = null;

		try {
			boolean isIPv4 = IPv4Packet.statIsIpv4Packet(headers);
			boolean isIPv6 = !isIPv4 && IPv6Packet.statIsIpv6Packet(headers);
			boolean isTCP = (isIPv4 || isIPv6) && IPFactory.isTCPPacket(headers);
			boolean isUDP = (isIPv4 || isIPv6) && !isTCP && IPFactory.isUDPPacket(headers);
			boolean isICMP = isIPv4 && !isTCP && !isUDP && IPv4Packet.getIpProtocolType(headers) == 1;
			boolean isICMPv6 = isIPv6 && !isTCP && !isUDP && IPv6Packet.getIpProtocolType(headers) == 58;
			
			if((isTCP || isUDP || isICMP || isICMPv6) && !(isIPv4 && isUDP &&
				new IPv4Packet(headers).getDestinationIPAsString().equals(ExporterService.getPrefCollectorAddr()))
			) {

				// IPv4
				if(isIPv4) {

					if(isTCP) {
						TemplateRecord template = getRegistredTemplateRecord(TPL_IPV4_TCP_ID);
						record = new RecordTCP(template);
						
					} else if(isUDP) {
						TemplateRecord template = getRegistredTemplateRecord(TPL_IPV4_UDP_ID);
						record = new RecordUDP(template);
						
					} else if(isICMP) {
						TemplateRecord template = getRegistredTemplateRecord(TPL_IPV4_ICMP_ID);
						record = new RecordICMP(template);
					}

					record.associateTemplateWithData(new FieldIPVersion(IPVersion.IPV4), false);
					record.associateTemplateWithData(new FieldIPv4AddrSrc(), true);
					record.associateTemplateWithData(new FieldIPv4AddrDst(), true);
					record.associateTemplateWithData(new FieldIPv4Length(), false);
				}

				// IPv6
				else if(isIPv6) {

					if(isTCP) {
						TemplateRecord template = getRegistredTemplateRecord(TPL_IPV6_TCP_ID);
						record = new RecordTCP(template);

					} else if(isUDP) {
						TemplateRecord template = getRegistredTemplateRecord(TPL_IPV6_UDP_ID);
						record = new RecordUDP(template);

					} else if(isICMPv6) {
						TemplateRecord template = getRegistredTemplateRecord(TPL_IPV6_ICMPV6_ID);
						record = new RecordICMPv6(template);
					}

					record.associateTemplateWithData(new FieldIPVersion(IPVersion.IPV6), false);
					record.associateTemplateWithData(new FieldIPv6AddrSrc(), true);
					record.associateTemplateWithData(new FieldIPv6AddrDst(), true);
					record.associateTemplateWithData(new FieldIPv6Length(), false);
				}

				// TCP
				if(isTCP) {
					record.associateTemplateWithData(new FieldIPProtocol(IPProtocol.TCP), false);
					record.associateTemplateWithData(new FieldTCPPortSrc(), true);
					record.associateTemplateWithData(new FieldTCPPortDst(), true);
					record.associateTemplateWithData(new FieldTCPFlags(), false);
					record.associateTemplateWithData(new FieldTCPAppName(this), false);
				}

				// UDP
				else if(isUDP) {
					record.associateTemplateWithData(new FieldIPProtocol(IPProtocol.UDP), false);
					record.associateTemplateWithData(new FieldUDPPortSrc(), true);
					record.associateTemplateWithData(new FieldUDPPortDst(), true);
					record.associateTemplateWithData(new FieldUDPAppName(this), false);
					
				// ICMP
				} else if(isICMP) {
					record.associateTemplateWithData(new FieldIPProtocol(IPProtocol.ICMP), false);	
				
				// ICMPv6
				} else if(isICMPv6) {
					record.associateTemplateWithData(new FieldIPProtocol(IPProtocol.IPV6_ICMP), false);	
				}
			}

			// Common
			if(record != null) {
				record.associateTemplateWithData(new FieldFirstSwitched(), false);
				record.associateTemplateWithData(new FieldLastSwitched(), false);
				record.associateTemplateWithData(new FieldPackets(), false);
				record.associateTemplateWithData(new FieldDeviceID(getContentResolver()), false);
				record.associateTemplateWithData(new FieldLocAlgo(42), false);
				record.associateTemplateWithData(new FieldLocTimestamp(locationFinder), false);
				record.associateTemplateWithData(new FieldLocLatInt(locationFinder), false);
				record.associateTemplateWithData(new FieldLocLatDec(locationFinder), false);
				record.associateTemplateWithData(new FieldLocLongInt(locationFinder), false);
				record.associateTemplateWithData(new FieldLocLongDec(locationFinder), false);
			}

			// Go! (or not)
			if(record != null) {
				processPacket(record, headers);
			}

		} catch(UnsupportedOperationException e) {
			return false;
		}
		
		return true;
	}
	
	@Override
	public void onStart() {
		connectivity = new Connectivity();
		registerReceiver(connectivity, new IntentFilter(ConnectivityManager.CONNECTIVITY_ACTION));
		Log.d("EXPORTER", "Connectivity class registred");
		
		locationFinder = LocationFinder.getInstance(this);
		locationFinder.start();
		Log.d("EXPORTER", "Location Finder is listening");
	}

	@Override
	public void onStop() {
		if(connectivity != null) {
			unregisterReceiver(connectivity);
			Log.d("EXPORTER", "Connectivity class unregistred");
		}
		
		if(locationFinder != null) {
			locationFinder.stop();
			Log.d("EXPORTER", "Location Finder is no longer listening");
		}
	}
}