package madynes.flowoid.exporterImpl;

import java.io.BufferedReader;
import java.io.IOException;
import java.io.InputStreamReader;
import java.nio.ByteBuffer;
import android.content.Context;
import android.content.pm.PackageInfo;
import android.content.pm.PackageManager.NameNotFoundException;
import madynes.flowoid.exporter.DataField;

public abstract class FieldAppName extends DataField {
	protected Context context;
	protected String value;

	public static enum AndroidUID {
		AID_RADIO         ("radio",1001),
		AID_BLUETOOTH     ("bluetooth", 1002),
		AID_GRAPHICS      ("graphics", 1003),
		AID_INPUT         ("input", 1004),
		AID_AUDIO         ("audio", 1005),
		AID_CAMERA        ("camera", 1006),
		AID_LOG           ("log", 1007),
		AID_COMPASS       ("compass", 1008),
		AID_MOUNT         ("mount", 1009),
		AID_WIFI          ("wifi", 1010),
		AID_ADB           ("adb", 1011),
		AID_INSTALL       ("install", 1012),
		AID_MEDIA         ("media", 1013),
		AID_DHCP          ("dhcp", 1014),
		AID_SDCARD_RW     ("sdcard_rw", 1015),
		AID_VPN           ("vpn", 1016),
		AID_KEYSTORE      ("keystore", 1017),
		AID_SHELL         ("shell", 2000),
		AID_CACHE         ("cache", 2001),
		AID_DIAG          ("diag", 2002),
		AID_NET_BT_ADMIN  ("net_bt_admin", 3001),
		AID_NET_BT        ("net_bt", 3002),
		AID_INET          ("inet", 3003),
		AID_NET_RAW       ("net_raw", 3004),
		AID_NET_ADMIN     ("net_admin", 3005),
		AID_MISC          ("misc", 9998),
		AID_NOBODY        ("nobody", 9999);

		private String name;
		private int value;
		
		AndroidUID(String name, int value) {
			this.name = name;
			this.value = value;
		}
		
		public String getName() {
			return name;
		}
		
		public int getValue() {
			return value;
		}
		
		public static String findByValue(int value) {
			for(AndroidUID aid : values()) {
				if(aid.getValue() == value) {
					return aid.getName();
				}
			}
			
			return null;
		}
	}
	
	public FieldAppName(Context context) {
		this.context = context;
	}

	/**
	 * Returns an Android app name from a L4 port.
	 * 
	 * @param port L4 port
	 * @param protocol IP protocol (e.g. tcp / tcp6)
	 * @return Application name
	 */
	protected String getAppName(int port, String protocol) {
		String appName = null;
		
		try {
			Process process = Runtime.getRuntime().exec("cat /proc/net/" + protocol);
			BufferedReader reader = new BufferedReader(new InputStreamReader(process.getInputStream()));
			String hexPort = String.format("%04x", port).toUpperCase();
			String line = reader.readLine();

			while(line != null && appName == null) {
				line = reader.readLine();
				
				if(line != null && line.matches(".*:" + hexPort + " .*")) {
					appName = getAppName(line);
				}
			}

			reader.close();
			process.waitFor();

		} catch (InterruptedException e) {
			e.printStackTrace();
		} catch (IOException e) {
			e.printStackTrace();
		} catch (NameNotFoundException e) {}
		
		return appName;
	}
	
	/**
	 * Returns an Android app name from a netstat line.
	 * 
	 * @param netstatLine Line extracted from /proc/net/proto.
	 * @return Application Name
	 * @throws NameNotFoundException When a NetFlow is not associated to an Android app
	 * (e.g. wget with uid 0)
	 */
	private String getAppName(String netstatLine) throws NameNotFoundException {
		String appName = null;
		String[] netstatLineChunks = netstatLine.split("\\s+");

		if(netstatLineChunks.length >= 9) {
			int uid = Integer.parseInt(netstatLineChunks[8]);
			
			appName = AndroidUID.findByValue(uid);
			
			if(appName == null) {
				String[] appNames = context.getPackageManager().getPackagesForUid(uid);
				
				if(appNames != null) {
					PackageInfo infos = context.getPackageManager().getPackageInfo(appNames[0], 0);
					
					if(infos != null) {
						appName = infos.packageName;
					}
				}
			}
			
			if(appName == null) {
				appName = "uid:" + uid;
			}
		}
		
		return appName;
	}

	@Override
	public ByteBuffer getValue() {
		ByteBuffer bytes = getByteBuffer();
		byte[] valueBytes = (value != null ? value : "Unknown").getBytes();
		
		for(int i = 0; i < valueBytes.length && i < bytes.capacity() - 1; i++) {
			bytes.put(valueBytes[i]);
		}
		
		bytes.putChar('\0');

		return bytes;
	}
}
