#!/usr/bin/perl

use strict;
use warnings;
use diagnostics;
use Data::Alias;
use Getopt::Long;
use Time::HiRes qw(gettimeofday tv_interval);
use DateTime::Format::DateParse;
 
my $optDebug = 0;
my $optDiscover = 0;
my $maxGapSequence = 3600/2;
my $optMinTopFreq = 1;
my $optMinVariationSameness = .6;
my $optMinCompressRate = .006;
my $optMinCount = 1;
my $optMinClusters = 10;
my $optStatTopTokens = 0;
my $optStatPatterns = 0;
my $optStatAvgVariations = 0;
my $optStatPrunedPatterns = 0;
my $optStatClusteredPatterns = 0;
my $optMinPatternSize = 10;

exit 1 unless GetOptions(
	"debug!" => \$optDebug,
	"discover!" => \$optDiscover,
	"maxGapSequence=s" => \$maxGapSequence,
	"minTopFreq=s" => \$optMinTopFreq,
	"minPatternSize=s" => \$optMinPatternSize,
	"minVariationSameness=s" => \$optMinVariationSameness,
	"minCompressRate=s" => \$optMinCompressRate,
	"minCount=s" => \$optMinCount,
	"minClusters=s" => \$optMinClusters,
	"statTopTokens" => \$optStatTopTokens,
	"statPatterns" => \$optStatPatterns,
	"statAvgVariations" => \$optStatAvgVariations,
	"statPrunedPatterns" => \$optStatPrunedPatterns,
	"statClusteredPatterns" => \$optStatClusteredPatterns
);

sub distance {
	my ($p_a, $p_b) = @_;

	alias my @a = @$p_a;
	alias my @b = @$p_b;

	my @distances;
	$distances[0] = [];

	for(my $j = 0; $j <= @b; $j++) {
		$distances[0][$j] = $j;
	}

	for(my $i = 1; $i <= @a; $i++) {
		$distances[$i] = [];
		$distances[$i][0] = $i;

		for(my $j = 1; $j <= @b; $j++) {
			$distances[$i][$j] = 0;
		}
	}

	for(my $i = 0; $i < @a; $i++) {
		for(my $j = 0; $j < @b; $j++) {
			if($a[$i] eq $b[$j]) {
				$distances[$i + 1][$j + 1] = $distances[$i][$j];
			} else {
				my $deletion = $distances[$i][$j + 1] + 1;
				my $insertion = $distances[$i + 1][$j] + 1;
				my $substitution = $distances[$i][$j] + 1;

				my $min = $deletion < $insertion ? $deletion : $insertion;
				$min = $substitution < $min ? $substitution : $min;

				$distances[$i + 1][$j + 1] = $min;
			}
		}
	}

	return $distances[@a][@b];
}

sub sameness {
	my ($p_a, $p_b) = @_;
	my $max = @$p_a > @$p_b ? @$p_a : @$p_b;

	return 1 - (distance($p_a, $p_b) / $max);
}

sub top {
	alias my @tokensSets = @_;
	my (@topSets, %topcounts, $previous);

	for(my $i = 0; $i < @tokensSets; $i++) {
		my @tokens = @{$tokensSets[$i]};
		$topSets[$i] = [];

		foreach my $p_token (@tokens) {
			$topcounts{@$p_token[0]}++;
		}

		$previous = -1;

		foreach my $p_token (@tokens) {
			if($topcounts{@$p_token[0]} > $optMinTopFreq) {
				push @{$topSets[$i]}, $p_token if $previous ne @$p_token[0];
				$previous = @$p_token[0];
			}
		}
	}

	return @topSets;
}

sub variations {
	my ($p_tokens, $p_variations, $p_counts, $setIteration) = @_;
	variationsIteration($p_tokens, $p_variations, $p_counts, $setIteration, $optMinPatternSize);
}

sub variationsIteration {
	my ($p_tokens, $p_variations, $p_counts, $setIteration, $iteration) = @_;

	print "\n-- Iteration #$setIteration-$iteration --\n\n" if $optDebug;

	alias my @tokens = @$p_tokens;
	alias my %variations = %$p_variations;
	alias my %counts = %$p_counts;

	my @changes;
	my $newGenerals = 0;

	for(my $i = 0; $i < @tokens - $iteration; $i++) {
		my $newgeneral = $tokens[$i][0];
		my @newgeneral_tokens = ($tokens[$i][0]);
		my $isVariation = 0;

		# New pattern
		for(my $j = $i + 1; $j < $i + $iteration + 1; $j++) {
			$newgeneral .= " $tokens[$j][0]";
			push @newgeneral_tokens, $tokens[$j][0];
		}

		if(exists $counts{$newgeneral}) {
			$counts{$newgeneral}++;

		} else {
			$counts{$newgeneral} = 1;

			# Find if it's a variation
			foreach my $general (keys %variations) {
				my @general_tokens = split /\s+/, $general;

				if(sameness(\@general_tokens, \@newgeneral_tokens) >= $optMinVariationSameness) {
					$isVariation = 1;

					push @{$variations{$general}}, $newgeneral;
					push @changes, $general unless $general ~~ @changes;

					#print "new variation: $general [$newgeneral]\n" if $optDebug;
					print "new variation\n" if $optDebug;
					last;
				}
			}

			unless($isVariation) {
				$variations{$newgeneral} = [];
				push @changes, $newgeneral;
				$newGenerals++;

				#print "new general: $newgeneral\n" if $optDebug;
				print "new general\n" if $optDebug;
			}
		}
	}

	foreach my $general (@changes) {

		my ($bestRepresentative, $bestRepresentativeIndex, $bestRepresentativeSizeGap);
		my $bestRepresentativeCompressRate = 0;

		my @patterns = ($general, @{$variations{$general}});
		my @chosenPatterns;
		my $sumSize = 0;

		foreach my $pattern (@patterns) {
			my @pattern_tokens = split /\s+/, $pattern;
			$sumSize += @pattern_tokens;
		}

		my $avgSize = int($sumSize / @patterns);

		foreach my $pattern (@patterns) {
			my $count = $counts{$pattern};
			my @pattern_tokens = split /\s+/, $pattern;
			my $sizeGap = abs(@pattern_tokens - $avgSize);
			my $compressRate = ($count * @pattern_tokens) / @tokens;

			if($compressRate >= $optMinCompressRate && $count >= $optMinCount) {
				my $i = (push @chosenPatterns, $pattern) - 1;

				if((!defined $bestRepresentativeSizeGap || $sizeGap <= $bestRepresentativeSizeGap) && $compressRate > $bestRepresentativeCompressRate) {
					$bestRepresentative = $pattern;
					$bestRepresentativeIndex = $i;
					$bestRepresentativeCompressRate = $compressRate;
					$bestRepresentativeSizeGap = $sizeGap;
				}
			}
		}

		delete $variations{$general};

		if(defined $bestRepresentative) {
			splice @chosenPatterns, $bestRepresentativeIndex, 1;
			$variations{$bestRepresentative} = \@chosenPatterns;
		} else {
			$newGenerals--;
			print "delete general\n" if $optDebug;
		}
	}

	if($newGenerals > 0) {
		variationsIteration(\@tokens, \%variations, \%counts, $setIteration, ++$iteration);
	}
}

sub pruning {
	alias my %variations = %{shift()};
	my @patternsBySize = sort { length $a <=> length $b } keys %variations;
	my @pruned;

	for(my $i = 0; $i < @patternsBySize; $i++) {
		my $pattern1 = $patternsBySize[$i];

		if(@{$variations{$pattern1}} > 0) {
			my $isMaximal = 1;

			for(my $j = @patternsBySize - 1; $j > $i && $j >= 0 && $isMaximal; $j--) {
				my $pattern2 = $patternsBySize[$j];
				$isMaximal = $pattern2 !~ /.*$pattern1.*/;
			}

			push @pruned, $pattern1 if $isMaximal;
		}
	}

	return @pruned;
}

sub clustering {
	alias my @patterns = @_;
	my (%distances, @centroids, @avgCentroids, @clusters, $centroidsChanges);
	my $minClusters = @patterns < $optMinClusters ? @patterns : $optMinClusters;
	my @patternsBySize = sort { length $a <=> length $b } @patterns;

	# Buffer distances
	foreach my $pattern1 (@patterns) {
		foreach my $pattern2 (@patterns) {
			my @pattern1_tokens = split /\s+/, $pattern1;
			my @pattern2_tokens = split /\s+/, $pattern2;

			$distances{$pattern1}{$pattern2} = distance(\@pattern1_tokens, \@pattern2_tokens);
		}
	}

	# Init centroids
	for(my $i = 0; $i < $minClusters; $i++) {
		my $unique = 0;
		$clusters[$i] = [];
		$centroids[$i] = $patternsBySize[$#patterns - $i];
		print "CENTROID #$i: $centroids[$i]\n" if $optDebug;
	}

	do {
		$centroidsChanges = 0;

		# Clustering
		foreach my $pattern (@patterns) {
			my $nearestCentroidIndex = 0;

			for(my $i = 1; $i < @centroids; $i++) {
				if($distances{$pattern}{$centroids[$i]} < $distances{$pattern}{$centroids[$nearestCentroidIndex]}) {
					$nearestCentroidIndex = $i;
				}
			}

			print "JOIN CENTROID #$nearestCentroidIndex: $pattern\n" if $optDebug;
			push @{$clusters[$nearestCentroidIndex]}, $pattern;
		}

		# New centroids
		for(my $i = 0; $i < @centroids; $i++) {
			my $minDistanceAvg;
			my $minDistanceAvgIndex;

			for(my $j = 0; $j < @{$clusters[$i]}; $j++) {
				my $distanceAvg;
				my $distanceSum = 0;

				foreach my $pattern (@{$clusters[$i]}) {
					$distanceSum += $distances{$clusters[$i][$j]}{$pattern}
				}

				$distanceAvg = $distanceSum / @{$clusters[$i]};

				if(!defined $minDistanceAvg || $distanceAvg < $minDistanceAvg) {
					$minDistanceAvg = $distanceAvg;
					$minDistanceAvgIndex = $j;
				}
			}

			if($centroids[$i] ne $clusters[$i][$minDistanceAvgIndex]) {
				$centroidsChanges = 1;
			}

			print "NEW CENTROID #$i: $clusters[$i][$minDistanceAvgIndex] (OLD: $centroids[$i])\n" if $optDebug;

			$centroids[$i] = $clusters[$i][$minDistanceAvgIndex];
			$clusters[$i] = [];
		}

	} while $centroidsChanges;

	return @centroids;
}

sub discover {
	my ($p_tokens, $p_patterns, $p_sequencesPatterns) = @_;

	alias my @tokens = @$p_tokens;
	alias my @sequencesPatterns = @$p_sequencesPatterns;
	my @patternsBySize = sort { length $a <=> length $b } @$p_patterns;
	my @maxPatternSize = split /\s+/, $patternsBySize[$#patternsBySize];

	print "discovering...\n" if $optDebug;

	for(my $i = 0; $i < @tokens;) {
		my $sequence_pattern = -1;
		my @sequence_tokens = ($tokens[$i][0]);
		my $sequence_last = $i;
		my $sequence_start = $tokens[$i][1];
		my $sequence_stop = $tokens[$i][2];
		my $add = 0;

		for($add = $i + 1; $add <= $i + (1 + 1 - $optMinVariationSameness) * @maxPatternSize && $sequence_pattern != $#patternsBySize && $add < @tokens; $add++) {
			push @sequence_tokens, $tokens[$add][0];

			for(my $k = 0; $k < @patternsBySize; $k++) {
				my @pattern_tokens = split /\s+/, $patternsBySize[$k];

				if(sameness(\@pattern_tokens, \@sequence_tokens) >= $optMinVariationSameness) {
					$sequence_pattern = $k;
					$sequence_last = $add;

					print "discovered\n" if $optDebug;
				}
			}
		}

		if($sequence_pattern != -1) {
			$sequence_stop = $tokens[$sequence_last][2];
			push @sequencesPatterns, [$sequence_pattern, $sequence_start, $sequence_stop];
			$i = $add;
		} else {
			$i++;
		}
	}
}

## MAIN

my (@tokens, @tokensSets, $nbTokens, %variations, %counts, @topTokens, @patterns, @prunedPatterns, @clusteredPatterns);
my ($t0, $t1, $t2, $t3, $t4, $statDurations);

my ($previous_end, $currentTokensSetId);
$currentTokensSetId = -1;

while (<STDIN>) {
	chomp;
	my ($startDate, $startTime, $duration, $src, $dst) = split /\s+/, $_;
	#my ($startDate, $startTime, $duration, $proto, $src, $void, $dst, $packets, $size) = split /\s+/;
	my ($srcIp, $srcPort) = $src =~ /^(.+):(\d+)$/;
	my ($dstIp, $dstPort) = $dst =~ /^(.+):(\d+)$/;

	if(defined $dstPort && $dstPort != 53 && $srcPort != 53 && $dstIp !~ /^10./) {
		my $start = DateTime::Format::DateParse->parse_datetime("$startDate $startTime");
		my $end = DateTime::Format::DateParse->parse_datetime("$startDate $startTime");
		$end->add(seconds => int($duration));

		push @tokens, [$dstIp, $start->epoch, $end->epoch];
	}
}

foreach my $p_token (sort { @$a[1] <=> @$b[1] } @tokens) {
	my @token = @$p_token;
	my $start = $token[1];
	my $end = $token[2];

	if(!defined $previous_end || $start - $previous_end > $maxGapSequence) {
		@{$tokensSets[++$currentTokensSetId]} = $p_token;
	} else {
		push @{$tokensSets[$currentTokensSetId]}, $p_token;
	}

	$previous_end = $end;
}

$t0 = [gettimeofday];
@topTokens = top @tokensSets;

$t1 = [gettimeofday];

for(my $i = 0; $i < @topTokens; $i++) {
	variations $topTokens[$i], \%variations, \%counts, $i;
}

@patterns = keys %variations;

$t2 = [gettimeofday];
@prunedPatterns = pruning \%variations;

print "\n-- CLUSTERING --\n\n" if $optDebug;

$t3 = [gettimeofday];
@clusteredPatterns = clustering @prunedPatterns;

$t4 = [gettimeofday];
$statDurations = tv_interval($t0, $t1) . '/' . tv_interval($t1, $t2) . '/' . tv_interval($t2, $t3) . '/' . tv_interval($t3, $t4);

# STATS

if($optStatAvgVariations) {
	my $nbVariations = 0;
	$nbVariations += @{$variations{$_}} foreach (keys %variations);
	print(($nbVariations / @patterns) . '/' . @patterns . " $statDurations\n");
}

print @patterns . " $statDurations\n" if $optStatPatterns;
print((1 - (@topTokens / $nbTokens)) . '/' . $nbTokens . " $statDurations\n") if $optStatTopTokens;
print((1 - (@prunedPatterns / @patterns)) . '/' . @patterns . " $statDurations\n") if $optStatPrunedPatterns;
print((1 - (@clusteredPatterns / @prunedPatterns)) . '/' . @prunedPatterns . " $statDurations\n") if $optStatClusteredPatterns;

# DEBUG

if($optDebug) {
	print "\n-- VARIATIONS --\n\n";

	foreach (sort { length $a <=> length $b } keys %variations) {
		print $_ . ': [' . join(' / ', @{$variations{$_}}) . "]\n";
	}

	print "\n-- PATTERNS --\n\n";
}

for(my $i = 0; $i < @clusteredPatterns; $i++) {
	print "[ #$i ] $clusteredPatterns[$i]\n\n";
}

if($optDiscover) {
	my @sequencesPatterns;

	print "-- DISCOVER --\n\n" if $optDebug;

	foreach my $p_tokens (@topTokens) {
		if(@clusteredPatterns > 0) {
			 discover $p_tokens, \@clusteredPatterns, \@sequencesPatterns;
		}
	}

	foreach my $sequence (@sequencesPatterns) {
		print DateTime->from_epoch(epoch => @$sequence[1])->set_time_zone("Europe/Paris") . " -> " . DateTime->from_epoch(epoch => @$sequence[2])->set_time_zone("Europe/Paris") . " : #" . @$sequence[0] . "\n";
	}
}

exit 0;
